"""
archinfo is a collection of classes that contain architecture-specific information.
It is useful for cross-architecture tools (such as pyvex).
"""
# pylint: disable=wildcard-import
from .arch import *
from .defines import defines
